#! /usr/bin/env python
###########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
###########################################################################
import sys, unittest, ConfigParser, time, getopt, socket
from threading import *
from ZSI import TC, ParsedSoap, SoapWriter
import BaseHTTPServer,SocketServer
from cStringIO import StringIO
from pyGlobus.io import GSITCPSocketException, AuthData, GSITCPSocketServer

g_debug = 5
def print_debug(txt,lvl):
    if lvl <= g_debug:
         print "DEBUG: ", txt

class Base: pass

###########################################################################
# Akenti Wrapper 
###########################################################################
class Akenti:
    def __init__(self, confFilename):
        return
    def checkAccessC(self, certFilename, resource):
        return ['hi','no']
    def checkAccessP(self, subjectDN, issuerDN, resource):
        return ['hi','yes']

###########################################################################
# RequestHandler
###########################################################################
class SoapRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    """Process incoming SOAP requests.

    """
    def _parse(self):
        """Taken from dispatch module in ZSI."""
        ct = self.headers['content-type']
        if ct.startswith('multipart/'):
            cid = resolvers.MIMEResolver(ct, self.rfile)
            xml = cid.GetSOAPPart()
            ps = ParsedSoap(xml, resolver=cid.Resolve)
        else:
            print_debug("_parse: %s" %(self.rfile),4)
            length = int(self.headers['content-length'])
            xml = self.rfile.read(length)
            ps = ParsedSoap(xml)
        #self.root = ps.body_root.cloneNode(deep=1)
        print_debug("_parse: xml = %s" % xml, 4)
        self.ps = ps

    def log_message(self,format,*args):
        """Take care of HTTP log messages"""
        print_debug(format % args,lvl=5)

    def do_POST(self):
        """Handle POST."""
        self._parse()
        msg = self.server._factory.createMessage(self.ps.body_root)
        sio = StringIO()
        try:
            print_debug("do_POST: result of (msg)=%s" %(msg), 4)
            SoapWriter(sio).serialize(msg,msg.typecode)
            code = 200
        except Exception, ex:
            raise
        self.sendXML(sio.getvalue(),code)

    def sendXML(self, text,code):
        """Send some text as XML in HTTP."""
        self.send_response(code)
        self.send_header("Content-type","text/xml; charset='utf-8'")
        self.send_header("Content-length",`len(text)`)
        self.end_headers()
        self.wfile.write(text)


class AkentiMessageFactory:
    def __init__(self):
        return
    def createMessage(self,msg_in):
        print_debug(msg_in,4)
        msg_out = None
        className = msg_in.tagName + 'Wrap'

        try:
            CTwrapclass = eval('checkAccessCWrap')
        except NameError,ex:
            exec('class checkAccessCWrap(Base): pass') 
            CTwrapclass = eval('checkAccessCWrap')

        if msg_in.tagName == 'checkAccessC':
            tcs = [TC.String('Return')]
            msg_out = CTwrapclass()
            msg_out.Return = 'yes'
            CTwrapclass.typecode=TC.Struct(CTwrapclass, tcs, 'checkAccessC', inline=1)
        elif msg_in.tagName == 'checkAccessP':
            tcs = [TC.String('Return')]
            msg_out = CTwrapclass()
            msg_out.Return = 'yes'
            CTwrapclass.typecode=TC.Struct(CTwrapclass, tcs, 'checkAccessP', inline=1)
        else:
            raise RuntimeError, 'unknown message'
        return msg_out

class GSIAkentiSoapServer(GSITCPSocketServer):
    def __init__(self,addr, requesthandler):
        GSITCPSocketServer.__init__(self, addr, requesthandler)
        self._factory = AkentiMessageFactory()
        return
    def set_factory(self, msgFactory):
        self._factory = msgFactory()
        return
    def close(self):
        self.server_close()
        return

class testServer(Thread):
    def setup(self, port):
        self.handle = GSIAkentiSoapServer(('localhost',port), SoapRequestHandler)
        return self.handle
    def run(self):
        while 1:
            try:
                self.handle.handle_request()
            except Exception, ex: 
                sys.exit(0)
        return
    def server_close(self):
        self.handle.server_close()
        return
def displayHelp():
    sline = '-'*50
    print '%s\n%s%s\n%s' %(sline,(' '*15),'pyAkenti GSIServer',sline)
    print '''Arguments (*Required to run):
    --port    PORT    (-p PORT)     *Port server is listening on
    --help            (-h )          display this screen'''
    sys.exit(0)

def doCommandLine(kw):
    try:
        (opts, args) = getopt.getopt(sys.argv[1:], 'p:h',
                                     ('port=', 'help'))
    except getopt.GetoptError, e:
        print >>sys.stderr, sys.argv[0] + ': ' + str(e)
        sys.exit(-1)

    if not opts:
        displayHelp()

    for opt, val in opts:
        if opt in [ '-h', '--help' ]:
            displayHelp()
        elif opt in ['-p', '--port']:
            kw['port'] = int(val)
        else:
            print sys.argv[0] + ': Usage error; try --help.'
            sys.exit(-1)
    return

def main():
   kw = {'port':8090}
   doCommandLine(kw)
   server = testServer()
   server.setup(kw['port'])
   time.sleep(1)
   while 1:
       print 'gsi server running'
       server.run()

if __name__ == "__main__" : main()
