from threading import Thread
from Queue import Queue

class Controller(Thread):
    """
    A controller class for registering events and listeners interested
    in those events.  Posting an event will cause all listeners
    registered for that event to be called, in the controller's
    thread.  Controllers are constructed by defautl to run as a
    daemon.
    """

    def __init__(self, daemon=1):
        """
        Initialize a Controller object.  To not run as a daemon,
        construct with daemon = 0.
        """
        Thread.__init__(self)
        self.setDaemon(daemon)
        self._eventQueue = Queue()
        self._eventMap = {}

    def registerListener(self, eventId, func):
        """
        Register the given listener for the given event.
        """
        l = self._eventMap.get(eventId, [])
        l.append(func)
        self._eventMap[eventId] = l

    def postEvent(self, event):
        """
        Fire the given event, all registered listners will be called,
        in the Controller's own thread.
        """
        self._eventQueue.put(event)

    def run(self):
        """
        Main controller event loop.  Check the eventQueue for events,
        when one appears execute the corresponding method according to
        eventMap.
        """
        print "Controller.run(): starting..."
        while 1:
            event = self._eventQueue.get(1)
            listeners = self._eventMap[event._id]
            for listener in listeners:
                apply(listener,(event,))



class Event:
    """
    A simple event base class containing and event id.
    """
    def __init__(self, id=None):
        """
        Construct and event with the given (defaults to None) id.
        """
        self._id = id
