############################################################################
# Keith Jackson, LBNL
# See Copyright for copyright notice!
###########################################################################
import ogsaBrowserController
from wxPython.wx import *

mWidgetCodes = {
    "ID_EXIT" : 1002,
}

g_control = None

class MyApp(wxApp):
    def OnInit(self):
        frame = childFrame(self, -1, "OGSA browser")
        frame.Show(true)
        self.SetTopWindow(frame)
        return true

class childFrame(wxFrame):
    def __init__(self, parent, ID, title):
        wxFrame.__init__(self, None, ID, title,
                                 wxDefaultPosition, wxSize(600, 600))
        self._parent = parent
        sizer = wxBoxSizer(wxHORIZONTAL)
        self._gui = GuiPanel(self)
        sizer.Add(self._gui, 1, wxEXPAND)
        self.MakeMenu()
        self.SetAutoLayout(true)
        self.SetSizer(sizer)
        defaultPanelColor(self)
        return
    
    def Quit(self, event):
        self.Close()

    def MakeMenu(self):
        fmenu = wxMenu()
        fmenu.Append(mWidgetCodes['ID_EXIT'], "&Exit", "Terminate the program")
        menuBar = wxMenuBar()
        menuBar.Append(fmenu,"&File")
        self.SetMenuBar(menuBar)
        EVT_MENU(self, mWidgetCodes['ID_EXIT'], self.Quit)

class GuiPanel(wxPanel):
    def __init__(self,parent):
        wxPanel.__init__(self,parent, -1)
        self._nb = NtBook(self)
        self._urlPanel = URLPanel(self)
        self._outputPanel = outputPanel(self)
        self.guiSetup()
        return

    def guiSetup(self):
        """ Do initializations here"""
        sz = wxBoxSizer(wxVERTICAL)
        sz.Add(self._urlPanel, 0, wxEXPAND)
        sz.Add(self._nb,1, wxEXPAND)
        sz.Add(self._outputPanel,1, wxEXPAND)
        self.SetAutoLayout(true)
        self.SetSizer(sz)
        self.Layout()
        

class outputPanel(wxPanel):
    def __init__(self, parent):
        wxPanel.__init__(self,parent, -1)
        self._outputPanel = wxTextCtrl(self,wxNewId(), style = wxTE_MULTILINE)
        self._title = wxStaticText(self, -1, ' Output  ')
        
        sz = wxBoxSizer(wxVERTICAL)
        sz1 = wxBoxSizer(wxHORIZONTAL)
        sz1.Add(self._title,1, wxALIGN_CENTER)
        sz.Add(sz1, 1, wxEXPAND)
        sz.Add(self._outputPanel,20, wxEXPAND)
            
        self.SetSizer(sz)
        self.SetAutoLayout(true)
        defaultPanelColor(self)

    def addText(self, newOutput):
        #Lose the old output for now
        self._outputPanel.SetValue(newOutput)     


class URLPanel(wxPanel):
    def __init__(self, parent):
        wxPanel.__init__(self, parent, wxNewId())
        urlTCtrl = wxTextCtrl(self, wxNewId(), "http://dummy.url", size=(50,-1))
        urlTCtrl.SetInsertionPoint(0)
        EVT_TEXT_ENTER(self, urlTCtrl.GetId(), self.onGo)
        self._urlTCtrl = urlTCtrl
        b = wxButton(self, wxNewId(), "Go")
        EVT_BUTTON(self, b.GetId(), self.onGo)
        self._goB = b
        sz = wxBoxSizer(wxHORIZONTAL)
        sz.Add(urlTCtrl, 1)
        sz.Add(b, 0)
        self.SetSizer(sz)
        self.SetAutoLayout(true)

    def onGo(self, event):
        ev = ogsaBrowserController.URLChangeEvent(self._urlTCtrl.GetLineText(0))
        print "str: %s" % self._urlTCtrl.GetLineText(0)
        g_control.postEvent(ev)
        
class NtBook(wxNotebook):
    def __init__(self, parent):
        wxNotebook.__init__(self, parent, wxNewId())
        pan = NBPanel(self)
        pan2 = NBPanel2(self)
        self.AddPage(pan, "FOO")
        self.AddPage(pan2, "BAR")

class NBPanel(wxPanel):
    def __init__(self, parent):
        wxPanel.__init__(self, parent, wxNewId(), style = wxRAISED_BORDER)
        win = wxWindow(self, -1, style = wxSIMPLE_BORDER)
        win.SetBackgroundColour(wxBLUE)
        self.win = win
        def OnCPSize(evt, win=win):
            win.SetSize(evt.GetSize())
        EVT_SIZE(self, OnCPSize)

class NBPanel2(wxPanel):
    def __init__(self, parent):
        wxPanel.__init__(self, parent, wxNewId(), style = wxRAISED_BORDER)
        win = wxWindow(self, -1, style = wxSIMPLE_BORDER)
        win.SetBackgroundColour(wxGREEN)
        self.win = win
        def OnCPSize(evt, win=win):
            win.SetSize(evt.GetSize())
        EVT_SIZE(self, OnCPSize)
        
def defaultPanelColor( win ):
    """
    Set default color for panels
    """
    win.SetBackgroundColour(wxWHITE)
    win.SetForegroundColour(wxBLACK)
    return

def main():
    global g_control
    g_control = ogsaBrowserController.OGSABrowserController()
    g_control.registerListener(ogsaBrowserController.URL_CHANGE,
                               ogsaBrowserController.onURLChange)
    g_control.start()
    app = MyApp(0)
    app.MainLoop()



if __name__ == "__main__": main()














