#! /usr/bin/env python
# $Header: /home/g5/repository/pyGridWare/pyGridWare/client.py,v 1.3 2003/03/19 02:11:53 boverhof Exp $
#
# Copyright (c) 2001 Zolera Systems.  All rights reserved.

from ZSI.client import *

class Binding_SOAPHEADER(Binding):
    """
    Object that represents a binding (connection) to a SOAP server.
    Once the binding is created, various ways of sending and
    receiving SOAP messages are available, including a "name overloading"
    style.

    ssl=2 Send method will attempt to import pyGlobus.io, if it is not
    available an ImportError will be thrown.
    """
    def __init__(self, nsdict=None, ssl=0, url=None, tracefile=None,
        host='localhost', readerclass=None, port=None,
        soapaction='"http://www.zolera.com"', **kw):
        Binding.__init__(self, nsdict, ssl, url, tracefile,\
                         host, readerclass, port, soapaction, **kw)
        self.soap_header = None

    def AddSOAPHeader(self, header):
        self.soap_header = header

    def accept_delegation(self, oids, buffers, time_req):
        """ Only valid for use when ssl=2
            Taken from pyGlobus.ZSISOAP
        """
        cred = self.socket.accept_delegation(oids, buffers, time_req)
        return cred

    def Send(self, url, opname, obj, nsdict=None, soapaction=None, **kw):
        '''Send a message.  If url is None, use the value from the
        constructor (else error). obj is the object (data) to send.
        Data may be described with a requesttypecode keyword, or a
        requestclass keyword; default is the class's typecode (if
        there is one), else Any.
        '''
        # Get the TC for the obj.
        if kw.has_key('requesttypecode'):
            print "if kw.has_key('requesttypecode'):"
            tc = kw['requesttypecode']
        elif kw.has_key('requestclass'):
            print "elif kw.has_key('requestclass'):"
            tc = kw['requestclass'].typecode
        elif type(obj) == types.InstanceType:
            print "elif type(obj) == types.InstanceType:"
            tc = getattr(obj.__class__, 'typecode')
            if tc == None: tc = TC.Any(opname, aslist=1)
        else:
            tc = TC.Any(opname, aslist=1)

        # Determine the SOAP auth element.
        if self.auth_style & AUTH.zsibasic:
            auth_header = _AuthHeader % (self.auth_user, self.auth_pass)
        else:
            auth_header = None

        # Serialize the object.
        s = StringIO.StringIO()
        d = {}
        if self.ns: 
            d[''] = self.ns
        d.update(nsdict or self.nsdict or {})

        if auth_header and self.soap_header:
            sw = SoapWriter(s, nsdict=d, header=(auth_header+self.soap_header))
        elif self.soap_header:
            sw = SoapWriter(s, nsdict=d, header=self.soap_header)
        else:
            sw = SoapWriter(s, nsdict=d, header=auth_header)

        if kw.has_key('_args'):
            sw.serialize(kw['_args'], tc)
        else:
            sw.serialize(obj, tc, typed=0)
        sw.close()
        soapdata = s.getvalue()

        # Tracing?
        if self.trace:
            print >>self.trace, "_" * 33, time.ctime(time.time()), "REQUEST:"
            print >>self.trace, soapdata

        # Send the request.
        if self.ssl == 0:
            self.h = httplib.HTTPConnection(self.host, self.port)
        elif self.ssl == 1:
            self.h = httplib.HTTPSConnection(self.host, self.port,
                        **self.ssl_files)
        elif self.ssl == 2:
            try:
                io
            except NameError, ex:
                from pyGlobus import io
            self.h = io.GSIHTTPConnection(self.host, self.port)
        else:
            raise RuntimeError, 'Binding does not know ssl=%d' %(self.ssl)


        self.h.connect()
        self.h.putrequest("POST", url or self.url)
        self.h.putheader("Content-length", "%d" % len(soapdata))
        self.h.putheader("Content-type", 'text/xml; charset=utf-8')
        self.h.putheader("SOAPAction", soapaction or self.soapaction)
        if self.auth_style & AUTH.httpbasic:
            val = _b64_encode(self.auth_user + ':' + self.auth_pass) \
                        .replace("\012", "")
            self.h.putheader('Authorization', 'Basic ' + val)
        for header,value in self.user_headers:
            self.h.putheader(header, value)
        self.h.endheaders()
        self.h.send(soapdata)

        # Clear prior receive state.
        self.data, self.ps = None, None

