############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
#     xml.xpath.ParsedNodeTest.QualifiedNameTest.match()
#  
###########################################################################
import types
from UserList import UserList
from UserDict import UserDict
from xml import xpath
from xml.xpath import Conversions
from xml.dom.ext import PrettyPrint, SplitQName
from xml.dom import Node
from pyGridWare.xmlInterface import XmlBase, FromDom, DomLookUp, DOMException
import threading
try:
    from ZSI.TC import XSD1
except ImportError, ex:
    XSD1 = 'http://www.w3.org/1999/XMLSchema'

WSDL   = 'http://schemas.xmlsoap.org/wsdl/'
STRICT = 0
DEBUG  = 0
RESERVED = 'Reserved Namespace'

def print_debug(msg, level=1, *l, **kw):
    if DEBUG >= level:
        for i in l:
            msg += '\n*\t' + str(i)
        for k,v in kw.items():
            msg += '\n**\t%s: %s' %(k,v)
        print "SCHEMA(%d): %s" %(level,msg)

        
class SchemaException(Exception): pass
class SingletonException(Exception): pass

###########################
# Markers
##########################
class SchemaComponent(XmlBase, FromDom):
    """
    """
    def __init__(self):
        self.name = None

    def _checkNode(self, tagName, node):
        if SplitQName(node.tagName)[1] != tagName:
            raise SchemaException, 'node<%s> is not <%s>' %(node.tagName, tagName)
       
    def fromDom(self, node):
	self.__setName(node)
	self.__setType(node)

    def __setName(self, node):
        self.name = xpath.Evaluate('string(@name)', node)

    def __setType(self, node):
        self.__qname = xpath.Evaluate('string(@type)', node)
        if self.__qname:
            self.__node = node
            tdc = TypeDescriptionComponent(self.__qname)
            tdc.fromDom(node)
            self.type = tdc
        else:
            self.type = None
    
class SchemaTypeEvaluator(DomLookUp):
    def __init__(self, node):
        DomLookUp.__init__(self)
        self.setSchemaNode(node)

    def setSchemaNode(self, node):
        DomLookUp.setSchemaNode(self, node)

    def getXmlns(self):
        return self.xmlns

    def getTypeDict(self, node, local=0):
        my_types = {}
	for my_type, type_list in DomLookUp.getTypeDict(self, node).items():
	    if my_type == 'element':
	        for child in type_list:
                    if local:
                        tp = LocalElementDeclaration(self)
                    else:
                        tp = Element(self)
                    tp.fromDom(child)
                    my_types[tp.name] = tp

	    elif my_type == 'complexType':
	        for child in type_list:
                    tp = ComplexType(self)
                    tp.fromDom(child)
                    my_types[tp.name] = tp

    	    elif my_type == 'simpleType':
	        for child in type_list:
                    tp = SimpleType()
                    tp.fromDom(child, self.targetNamespace, **self.xmlns)
                    my_types[tp.name] = tp

	    elif my_type == 'any':
	        for child in type_list:
                    tp = WildCardDeclaration(self)
                    tp.fromDom(child)
                    my_types[tp.name] = tp

	    else:
	        raise SchemaException, 'Unknown type (%s)' %(my_type)
	
	return my_types


class Schema(SchemaComponent):
    """
        xmlns,targetNamespace in <schema>,
        you can dump the node in the constructor and get
        Schema.targetNamespace

    """
    def __init__(self):
        self.types = {}
        self.targetNamespace = None
       
    def setTargetNamespace(self, node):
        self.targetNamespace = xpath.Evaluate('string(@targetNamespace)', node)

    def fromDom(self, node):
        schemaEval = SchemaTypeEvaluator(node)
	self.xmlns = schemaEval.getXmlns()
	if not self.targetNamespace:
            self.setTargetNamespace(node)
        self.types.update(schemaEval.getTypeDict(node))


#
#  Element, Attribute
#
class AnyBase:
    """ 
        processContents - skip, lax, strict
        namespace       - [list of allowable ns,
	                   or classes of namespaces for any content]
    """
    def __init__(self):
        raise NotImplementedError, 'AnyBase constructor is not to be called (%s)' %(self.__class__)

class Attribute(SchemaComponent):
    def __init__(self):
	SchemaComponent.__init__(self)
        self.name = None 
        self.type = None 
        self.use = None 

    def __setUse(self, node):
        self.use = xpath.Evaluate('string(@use)', node)

    def fromDom(self, node):
	SchemaComponent.fromDom(self, node)
        self.__setUse(node)


class Element(SchemaComponent):
    """
        name
	type
    """
    def __init__(self, schemaEval):
	SchemaComponent.__init__(self)
	self.name = None
	self.type = None
	self.targetNamespace = schemaEval.targetNamespace
	self._schemaEval = schemaEval

    def fromDom(self, node):
	SchemaComponent.fromDom(self, node)
	if not isinstance(self.type, TypeDescriptionComponent) and not self.type:
	    types_d = self._schemaEval.getTypeDict(node)
	    if len(types_d) == 1:
                self.type = types_d.values()[0]
	    #else:
	    #	raise SchemaException, 'element is missing a type'

class LocalElementDeclaration(Element):
    def __init__(self, schemaEval):
        Element.__init__(self, schemaEval)
        self.minOccurs = '1'
        self.maxOccurs = '1'

    def fromDom(self, node):
        self.minOccurs = xpath.Evaluate('string(@minOccurs)', node) or self.minOccurs
        self.maxOccurs = xpath.Evaluate('string(@maxOccurs)', node) or self.maxOccurs
        Element.fromDom(self, node)
    
class WildCardDeclaration(LocalElementDeclaration, AnyBase):
    """
        namespace 
    """
    def __init__(self, schemaEval):
        LocalElementDeclaration.__init__(self, schemaEval)
	return

    def fromDom(self, node):
	self.name = node.tagName
        LocalElementDeclaration.fromDom(self, node)
        self.namespace = xpath.Evaluate('string(@namespace)', node)
        seqNode = xpath.Evaluate('targetNamespace', node)


class AnyAttribute(Attribute, AnyBase):
    def fromDom(self, node): pass



class AnyType(SchemaComponent):
    """
        fromDom node is a <element>, <complexType>, or <simpleType>
        self.type is not resolved until its first use
    """
    type_tags = {'element':0, 'simpleType':1, 'complexType':2}
    def __setTagName(self, node):
        if AnyType.type_tags.get(SplitQName(node.tagName)[1]) == None:
            raise SchemaException, 'node (%s) is not a valid Type %s'\
                  %(node.tagName, Type.type_tags.keys())
        self._tagName = node.tagName

    def fromDom(self, node):
	SchemaComponent.fromDom(self, node)
        self.__setTagName(node)


class ComplexType(AnyType):
    """ <sequence> - subset of elements
        <choice>   - choose one element
        <all>      - 
        <attribute>- 
    """
    def __init__(self, schemaEval):
	self._schemaEval = schemaEval
	self.targetNamespace = schemaEval.targetNamespace
        self.sequences = self.Sequence(schemaEval)
        self.choices = []
        self.alls = self.All(schemaEval)
        self.attributes = []
        return

    def fromDom(self, node):
	AnyType.fromDom(self, node)

	"""HACK: move to xmlInterface, use Context
	"""
	for child in node.childNodes:
	    if child.nodeType == Node.ELEMENT_NODE:
	        if SplitQName(child.tagName)[1] == 'sequence':
                   self.sequences.append(self.Sequence(self._schemaEval))
                   self.sequences[-1].fromDom(child)
                   
                elif SplitQName(child.tagName)[1] == 'choice':
                    self.choices.append(self.Choice(self._schemaEval))
                    self.choices[-1].fromDom(child)
            
                elif SplitQName(child.tagName)[1] == 'all':
                    self.alls.append(self.All(self._schemaEval))
                    self.alls[-1].fromDom(child)
                    
                elif SplitQName(child.tagName)[1] == 'attribute':
                   self.attributes.append(Attribute())
                   self.attributes[-1].fromDom(child)

    class Sequence(UserList):
	def __init__(self, schemaEval):
             UserList.__init__(self)
	     self._schemaEval = schemaEval
             self.minOccurs = '1'
             self.maxOccurs = '1'

	def fromDom(self, node):
             self.minOccurs = xpath.Evaluate('string(@minOccurs)', node) or self.minOccurs
             self.maxOccurs = xpath.Evaluate('string(@maxOccurs)', node) or self.maxOccurs
             
             d = self._schemaEval.getTypeDict(node, local=1)
	     for i in d.values(): self.append(i)

    class All(UserList):
	def __init__(self, schemaEval):
             UserList.__init__(self)
	     self._schemaEval = schemaEval

	def fromDom(self, node):
             d = self._schemaEval.getTypeDict(node, local=1)
	     for i in d.values(): self.append(i)

    class Choice(UserList):
	def __init__(self, schemaEval):
             UserList.__init__(self)
	     self._schemaEval = schemaEval
             self.minOccurs = '1'
             self.maxOccurs = '1'

	def fromDom(self, node):
             self.minOccurs = xpath.Evaluate('string(@minOccurs)', node) or self.minOccurs
             self.maxOccurs = xpath.Evaluate('string(@maxOccurs)', node) or self.maxOccurs
             d = self._schemaEval.getTypeDict(node, local=1)
	     for i in d.values(): self.append(i)


class SimpleType(AnyType):
    def __init__(self):
        self.restriction = None
        self.qname = None
        self.name = None
	self.targetNamespace = None
        self.final = None

    def fromQName(self, qname, targetNamespace, **xmlns):
        self.qname = qname
	qname = SplitQName(qname)
	if qname[0] and xmlns:
	     self.targetNamespace = xmlns[qname[0]]
	else:
	     self.targetNamespace = targetNamespace
            
    def fromDom(self, node, targetNamespace, **xmlns):
	AnyType.fromDom(self, node)
        self.final = xpath.Evaluate('string(@final)', node) or None
	for child in node.childNodes:
	    if child.nodeType == Node.ELEMENT_NODE:
	        if SplitQName(child.tagName)[1] == 'restriction':
                   self.restriction = self.Restriction()
                   self.restriction.fromDom(child, targetNamespace, **xmlns)
                   break
	        elif SplitQName(child.tagName)[1] == 'list':
                   self.restriction = self.Restriction()
                   self.restriction.fromDom(child, targetNamespace, **xmlns)
                   break
	        elif SplitQName(child.tagName)[1] == 'union':
                   self.restriction = self.Restriction()
                   self.restriction.fromDom(child, targetNamespace, **xmlns)
                   break

    class Restriction:
	def fromDom(self, node, targetNamespace, **xmlns):
	    for a in node.attributes.values(): 
                 if a.name == 'base':
                     self.base = TypeDescriptionComponent(a.value)
                     self.base.fromDom(node)
                 else:
                     setattr(self, a.name, a.value)
                  
            """
	    if SplitQName(a.value)[0]:
	         self.targetNamespace = self.schema.xmlns.get(SplitQName(a.value)[0])
	    else:
	         self.targetNamespace = targetNamespace
	    self.default_xmlns = xmlns.get('xmlns', None)
            """

            #tdc = TypeDescriptionComponent(node)

class AnonymousType(AnyType, UserList):
    """ xs:element   - ComplexType
        xs:attribute - SimpleType

	These can contain type structure w/o a name

    """
    def __init__(self, schemaEval):
	self._schemaEval = schemaEval

    def fromDom(self, node):
         d = self._schemaEval.fromDom(self, node)
	 for i in d.values(): self.append(i)


###########################
###########################
class TypeDescriptionComponent(XmlBase):
    """
        This object calls specific methods on self._type object
        so it appears to the user as one in the same.
    
        tns_schema
          - class variable mappings of targetNamespace to [schema,].
            ns scope limits tns-schema mapping to document.
        __qname
          - qname of type, local to the definition document.
        fromDom
          - takes the local definition dom node, in which qname is
            interpretable.  Save tns:Schema() mapping.
    """
    tns_schema = {}
    def __init__(self, qname):
        self.qname = qname
        self.targetNamespace = None
        self.__schema = None
        return
    
    def __getattr__(self, cmd):
        if cmd == 'type' or 'base':
            self.__evalSchema()
            if isinstance(self.type, SimpleType) or\
               isinstance(self.type, ComplexType) or\
               isinstance(self.type, Element):
	        return eval('self.' + cmd)
            else:
                raise SchemaException, 'Error type = %s' %self.type

    def __nonzero__(self):
	""" This is needed for if statements when __getattr__ is defined, although
	    this func is not passed to __getattr__??
	"""
        return 1
    
    def fromDom(self, node):
        lookup = DomLookUp()
        prefix,name = SplitQName(self.qname)
        lookup.setSchemaDefNode(node)

        if prefix and lookup.xmlns.has_key(prefix):
            self.targetNamespace = lookup.xmlns[prefix]
        elif prefix:
            lookup.setDefinitionsNode(node)
            self.targetNamespace = lookup.xmlns[prefix]
        else:
            self.targetNamespace = lookup.targetNamespace

        self.xmlns = lookup.xmlns
        return

    def __evalSchema(self):
        prefix,name = SplitQName(self.qname)

        if prefix and self.xmlns.has_key(prefix):
            schemas = TypeDescriptionComponent.tns_schema.get(self.xmlns[prefix])
        else:
            schemas = TypeDescriptionComponent.tns_schema.get(self.targetNamespace)

        if schemas and len(schemas) > 1:
            raise SchemaException,\
                  'no local scoping of multiple schema into a targetNamespace for multiple documents'
        elif schemas and schemas[0].types.has_key(name):
            self.__schema = schemas[0]
            self.__schema.types
            self.type = schemas[0].types[name]
        else:
            self.type = SimpleType()
            if prefix and self.xmlns.has_key(prefix):
                self.type.fromQName(qname=self.qname, targetNamespace=self.xmlns[prefix])
            else:
                self.type.fromQName(qname=self.qname, targetNamespace=self.xmlns['xmlns'])
        return
