############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import types, sys, copy, urlparse
from wsdl import Wsdl,\
     Binding, BFault, BInput, BOperation, BOutput,\
     PortType, Port, Operation, Output, Input,\
     MessageDescriptionComponent, Message, Part,\
     SoapAddress, SoapBinding, SoapBindingContainer,\
     SoapBindingElement, SoapBody, SoapOperation
from schema import AnyType, ComplexType, \
     SimpleType, Element, WildCardDeclaration, TypeDescriptionComponent

from client import Binding_SOAPHEADER as Binding
from typecodes import BaseTypeInterpreter, TC_Base, TC_Template, TC_NOTYPE
from typecodes import TC_Struct as Struct
from ZSI import TC, fault
from xml.dom.ext import SplitQName


class Service: pass

class Stub:
    """
    Parent class of all stub classes.
    key words dict is for ZSI Binding object
   
    class variables:
      tc_choice   - {'operation_name':ZSIResponseHandler}

    kw['url'] - If specified it overrides a soapAction element
      in ZSIMessage specified in wsdl.
     
    """
    def __init__(self, nsdict={}, **kw):
        self.kw = kw
        self.soap_header = None
        self.nsdict = nsdict
        return
    
    """
    def __getKW(self, **bindings):
	kw = {}
	kw.update(self.kw)
	if not self.kw.has_key('url') \
	   and bindings.has_key('soap_action'):
	    kw['url'] = bindings['soap_action']
	if bindings.has_key('soap_address'):
	    scheme, netloc = urlparse.urlparse(bindings['soap_address'])[0:2]
	    netloc = netloc.split(':') + [80,]
	    kw['host'] = netloc[0]
	    kw['port'] = int(netloc[1])
	    if scheme != urlparse.urlparse(bindings.get('soap_transport'))[0]:
		raise RuntimeError, 'soap address scheme does not match transport scheme'
        return kw
    """

    def add_soap_header(self, header):
        self.soap_header = header
        return

    def RPCFault(self, cmd, request, responseTC, *faults, **kw):
	""" Two-way communication, 
	"""
        kwc = self.kw.copy()
        kwc.update(kw)
	client = Binding(nsdict=self.nsdict, **kwc)
	if self.soap_header:
	    client.AddSOAPHeader(self.soap_header)

	client.Send(None, cmd, request)
	if client.IsAFault():
            print "FAULT PS: ", client.ps
            msg = fault.FaultFromFaultMessage(client.ps)
            if msg.detail and faults:
                print "DETAIL: ", msg
                print "DETAIL: ", faults
                client.ps.Parse(faults[0])

        elif responseTC:
	    msg = client.ps.Parse(responseTC)
        else:
	    raise RuntimeError, 'No typecode to parse response msg'
        return msg 

    def RPC(self, request, response=None, **kw):
	""" Two-way communication
	"""
        kwc = self.kw.copy()
        kwc.update(kw)
	client = Binding(nsdict=self.nsdict, **kwc)
	if self.soap_header:
	    client.AddSOAPHeader(self.soap_header)

	return client.RPC(None, None, request, response)

    def Send(self, request, **kw):
	""" One-way communication, Send Only
	"""
        kwc = self.kw.copy()
        kwc.update(kw)
	client = Binding(nsdict=self.nsdict, **kwc)
	if self.soap_header:
	    client.AddSOAPHeader(self.soap_header)
	client.Send(None, None, request)
        return

    def Receive(self, response, **kw):
	""" One-way communication, Receive Only
	"""
        kwc = self.kw.copy()
        kwc.update(kw)
	client = Binding(nsdict=self.nsdict, **kwc)
	return client.Receive(response)
"""
    class __Message:
        def __init__(self, call, cmd):
            self.__call = call
            self.__cmd  = cmd
	    return

        def __call__(self, request):
            return self.__call(self.__cmd, request)
"""

"""
class ZSIMessage:
    '''
         name - Name of request operation
    '''
    #def __init__(self, name, response=None):
    def __init__(self):
	self._instance = None

    def _getInstance(self):
        return self._instance

    def _getResponse(self):
        #return self.__class__.responseclass
        return self.__class__.responseclass.responseclass()

class ZSIRequestMessage(TC.TypeCode):  pass

class ZSIResponseMessage(TC.TypeCode):
    def __init__(self):
	TC.TypeCode.__init__(self)
"""




