############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import ConfigParser, unittest
from pyGridWare import typecodes

from ZSI import SoapWriter, ParsedSoap
from ZSI.TC import String, Iint, Any
from ZSI.TCcompound import Choice, Array
from ZSI.TCtimes import gDateTime
from pyGridWare.typecodes import TC_Struct, ComplexType, Sequence, Choice, All
import sys, time, copy, types, StringIO

"""
Unittest for the typecodes
"""

#The config file is ../../test/config.py
config = None # ConfigParser object

class myChoice:
    def __init__(self): pass
myChoice.definition = [Choice(myChoice, [String('myString'), gDateTime('myDate')], 
                       pname='myChoice', 
                       maxOccurs=3, 
                       minOccurs=1)]

class myChoice2:
    def __init__(self): pass
myChoice2.definition = [Choice(myChoice, [String('myString2'), gDateTime('myDate2')], 
                        pname='myChoice2', 
                        maxOccurs=3, 
                        minOccurs=1)]


class mySequence:
    def __init__(self): pass
mySequence.definition = [Sequence(mySequence, [String('one', optional=1), String('two')], 
                         pname='mySequence', 
                         maxOccurs=4, 
                         minOccurs=1)]

class cmplxType:
    def __init__(self): pass
cmplxType.typecode = TC_Struct(cmplxType, myChoice.definition, "myStruct")

class cmplxType2:
    def __init__(self): pass
cmplxType2.typecode = ComplexType(cmplxType, myChoice.definition+myChoice2.definition, "myStruct")


class cmplxTypeS:
    def __init__(self): pass
cmplxTypeS.typecode = ComplexType(cmplxTypeS, mySequence.definition, "myStruct")


class cmplxTypeCS:
    def __init__(self): pass
cmplxTypeCS.typecode = ComplexType(cmplxTypeCS, myChoice.definition+mySequence.definition, "CS")

class cmplxTypeSCSC:
    def __init__(self): pass
cmplxTypeSCSC.typecode = ComplexType(cmplxTypeSCSC, 
                                     (mySequence.definition +
                                      myChoice.definition   +
                                      mySequence.definition +
                                      myChoice.definition),
                                     "SCSC")

# (mySequence.definition + myChoice.definition), 
class holder: pass


class TypecodesTest(unittest.TestCase):
    "Test case for wsdl.Wsdl"

    def setUp(self):
        print "*"*50
        return

    def tearDown(self):
        return

    def checkChoice(self):
	request = cmplxType2()
	request.content  = []
	request.content.append(holder())
	request.content[0].myString = ['first', 'second']
	request.content[0].myDate   = [time.gmtime(1049840900)]

	request.content.append(holder())
	request.content[1].myString2 = 'what'
	request.content[1].myDate2    = [time.gmtime(1049850900), time.gmtime(1049851900)]
        content = copy.deepcopy(request.content)
       
	soapdata = getSoapData(request)
	ps = ParsedSoap(soapdata)
	data = ps.Parse(request.typecode)

        print "%s\n\n%s" %('='*50,soapdata)
        for k,v in data.__dict__.items(): print "%s = %s" %(k,v)
        print "\n%s" %('='*50)

        request.content = content
        #check data
	self.failUnless(request.content[0].myString[0] == data.myString[0], 'failed')
	self.failUnless(request.content[0].myString[1] == data.myString[1], 'failed')
	self.failUnless(request.content[1].myString2 == data.myString2[0], 'failed')

    def checkSequence(self):
	request = cmplxTypeS()
	request.content  = []
	request.content.append(holder())
	request.content[0].one = ['firstS']
	request.content[0].two = ['firstS1', 'secondS2', 'thirdS3']
        
        soapdata = getSoapData(request)
        print "%s\n\n%s" %('='*50,soapdata)

            
	ps = ParsedSoap(soapdata)
	data = ps.Parse(request.typecode)

        for k,v in data.__dict__.items(): print "%s = %s" %(k,v)
        print "\n%s" %('='*50)

        #check data
	self.failUnless(request.content[0].one[0] == data.one[0], 'failed')

	self.failUnless(request.content[0].two[0] == data.two[0], 'failed')
	self.failUnless(request.content[0].two[1] == data.two[1], 'failed')
	self.failUnless(request.content[0].two[2] == data.two[2], 'failed')
   

    def checkCS(self):
	request = cmplxTypeCS()
	request.content  = []

	request.content.append(holder())
	request.content[0].myString = ['choice01', 'choice02']
	request.content[0].myDate    = [time.gmtime(1049840900)]

	request.content.append(holder())
	request.content[1].one = ['seq11', 'seq12']
	request.content[1].two = ['seq21', 'seq22']
        content = copy.deepcopy(request.content)

	soapdata = getSoapData(request)
        print "%s\n\n%s" %('='*50,soapdata)

	ps = ParsedSoap(soapdata)
	data = ps.Parse(request.typecode)

        for k,v in data.__dict__.items(): print "%s = %s" %(k,v)
        print "\n%s" %('='*50)

        request.content = content
        #check data
	self.failUnless(request.content[0].myString[0] == data.myString[0], 'failed')
	self.failUnless(request.content[0].myString[1] == data.myString[1], 'failed')
	self.failUnless(types.TupleType == type(data.myDate[0]), 'failed')

	self.failUnless(request.content[1].one[0] == data.one[0], 'failed')
	self.failUnless(request.content[1].one[1] == data.one[1], 'failed')

	self.failUnless(request.content[1].two[0] == data.two[0], 'failed')
	self.failUnless(request.content[1].two[1] == data.two[1], 'failed')

 
    def checkSCSC(self):
	request = cmplxTypeSCSC()
	request.content  = []

	request.content.append(holder())
	request.content[0].one = ['seq01']
	request.content[0].two = ['seq02']

	request.content.append(holder())
	request.content[1].myString = ['choice11']
	request.content[1].myDate    = [time.gmtime(1049840900)]

	request.content.append(holder())
	request.content[2].one = ['seq21', 'seq22']
	request.content[2].two = ['seq21', 'seq22']

	request.content.append(holder())
	request.content[3].myString = ['choice31']
	request.content[3].myDate    = [time.gmtime(1049840900), time.gmtime(1059840900)]
        content = copy.deepcopy(request.content)

	soapdata = getSoapData(request)
        print "%s\n\n%s" %('='*50,soapdata)

	ps = ParsedSoap(soapdata)
	data = ps.Parse(request.typecode)

        for k,v in data.__dict__.items(): print "%s = %s" %(k,v)
        print "\n%s" %('='*50)

        request.content = content
	self.failUnless(request.content[0].one[0] == data.one[0], 'failed')
	self.failUnless(request.content[0].two[0] == data.two[0], 'failed')

	self.failUnless(request.content[1].myString[0] == data.myString[0], 'failed')
	self.failUnless(types.TupleType == type(data.myDate[0]), 'failed')

	self.failUnless(request.content[2].one[0] == data.one[1], 'failed')
	self.failUnless(request.content[2].one[1] == data.one[2], 'failed')
	self.failUnless(request.content[2].two[0] == data.two[1], 'failed')
	self.failUnless(request.content[2].two[1] == data.two[2], 'failed')

	self.failUnless(request.content[3].myString[0] == data.myString[1], 'failed')
	self.failUnless(types.TupleType == type(data.myDate[0]), 'failed')
	self.failUnless(types.TupleType == type(data.myDate[1]), 'failed')


def makeTestSuite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TypecodesTest, "check"))
    return suite

def main():
    unittest.main(defaultTest="makeTestSuite")
                  
def getSoapData(pyobj):
    tc = getattr(pyobj.__class__, 'typecode')
    s = StringIO.StringIO()
    sw = SoapWriter(s, nsdict={})
    sw.serialize(pyobj, tc, typed=0)
    sw.close()
    soapdata = s.getvalue()
    return soapdata


if __name__ == "__main__" : main()
