#a###########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
#     xml.xpath.ParsedNodeTest.QualifiedNameTest.match()
#  
###########################################################################
import urlparse, cStringIO, types
from UserList import UserList
from Ft.Lib import pDomlette
#from xml.xpath import Conversions
from xml import xpath
from xml.dom.ext import SplitQName
#from xml.dom.ext import PrettyPrint, SplitQName
from xml.dom import Node
from pyGridWare.xmlInterface import XmlBase, FromDom, XMLNS, DomLookUp
from pyGridWare.schema import Schema, TypeDescriptionComponent

STRICT = 0
DEBUG  = 0

def print_debug(msg, level=1, *l, **kw):
    if DEBUG >= level:
        for i in l:
            msg += '\n*\t' + str(i)
        for k,v in kw.items():
            msg += '\n**\t%s: %s' %(k,v)
        print "WSDL(%d): %s" %(level,msg)

        
def addSchemaToTypesDescriptionComponent(schema):
    """ TypeDescriptionComponent is only class that
        needs access to the schema.
        For now place it in a class variable.  TDC shouldn't
        need to know anything about the schema class.  It
        only contains a qname and a node.
    """
    print_debug("__addSchemaToTypesDescriptionComponent %s" %(schema), 4)
    if TypeDescriptionComponent.tns_schema.get(schema.targetNamespace):
        print_debug('**1st %s' %(schema.targetNamespace), 4)
        TypeDescriptionComponent.tns_schema[schema.targetNamespace].append(schema)
    elif schema.targetNamespace:
        print_debug('**2nd %s' %(schema.targetNamespace), 4)
        TypeDescriptionComponent.tns_schema[schema.targetNamespace] = [schema]
    else:
        raise WsdlException, 'Schema has no targetNamespace'
    return

######################################################
# Wsdl
# pluggable xml parser??
#    file:/usr/local/java/ogsa-tp4/schema/samples/google/google_search_service.wsdl
#    /usr/local/java/ogsa-tp4/schema/samples/google/google_search_service.wsdl
####################################################
class WsdlException(Exception): pass
class SingletonException(Exception): pass
class Wsdl(XmlBase, FromDom):
    """ imported doms
        __wsdl_doms  -- .wsdl <definitions>, level {url:dom}
        __xsd_doms   -- .xsd <schema>
        __top_dom    -- hierarchical
    """
    singleton = 0
    def __init__(self, url):
        self.x = None
        if Wsdl.singleton:
            raise SingletonException, 'only one instance of Wsdl can exist at a time'
        self.services = []
        self.schemas = {}
        self.__wsdl_doms = {}
        self.__xsd_doms = {}
        reader = pDomlette.PyExpatReader()
        node = self.__readFromUrl(reader, url)
        del reader
        
	self.fromDom(node)

    def __del__(self):
	""" HACK, need to reset this class variable in order to 
	    create a new Wsdl object
	"""
        TypeDescriptionComponent.tns_schema = {}
        Wsdl.singleton = 0
        del self.__wsdl_doms, self.__xsd_doms, self.__top_dom

    def fromDom(self, node):
	""" first read in schema from .xsd files, .wsdl files
	    then read in service descriptions.
	"""
        self.__top_dom = node
        self.name = xpath.Evaluate('string(@name)', node.documentElement)
        for dom in self.__xsd_doms.values():
            self.__readSchemaFromDom(dom.documentElement)
        for tns, dom in self.__wsdl_doms.items():
            self.__readTypesFromDom(dom.documentElement)
        for dom in self.__wsdl_doms.values():    
            self.__readServicesFromDom(dom.documentElement)
        return

    def __readSchemaFromDom(self, node):
        """ Check targetNamepace of node to see if schema already exists.
	    Read in <schema>'s from .xsd
        """
        schema = Schema()
        schema.setTargetNamespace(node)

        if not TypeDescriptionComponent.tns_schema.get(schema.targetNamespace):
            schema.fromDom(node)
            addSchemaToTypesDescriptionComponent(schema)
        self.schemas[schema.targetNamespace] = schema
        return

    def __readTypesFromDom(self, node):
        """ Read in <types>'s from .wsdl
        """
        lookup = DomLookUp()
        lookup.setDefinitionsNode(node)
        nodes = lookup.getTypes()
        nsd = lookup.getImportNS()

        for n in nodes:
            types = Types()
            types.fromDom(n)
            for prefix,uri in lookup.xmlns.items():
                if not types.schema.xmlns.has_key(prefix):
                    types.schema.xmlns[prefix] = uri

            addSchemaToTypesDescriptionComponent(types.schema)
            self.schemas[types.schema.targetNamespace] = types.schema
        return
    
    def __readServicesFromDom(self, node):
        """ Read in <service>'s from .wsdl
        """
        lookup = DomLookUp()
        lookup.setDefinitionsNode(node)
        nodes = lookup.getService()

        for i in range(len(nodes)):
            self.services.append(Service())
            self.services[i].fromDom(nodes[i])
        return

    def __readFromUrl(self, reader, uri, base_uri=''):
        """ condensed urls for comparison w/ new imports
	    definitions.documentElement: schema, definitions
            check import.namespace == [definitions,schema].targetNamespace
        """
        my_url = Url(base_uri+uri)
        for dom_dict in [self.__wsdl_doms, self.__xsd_doms]:
            if dom_dict.has_key(str(my_url)):
                dom = dom_dict[str(my_url)]
                print_debug("FOUND (%s)  %s " %(dom.documentElement, uri))
                break
        else:
            dom = reader.fromUri(str(my_url))
            if str(my_url).endswith('.wsdl'):
                self.__wsdl_doms[str(my_url)] = dom
            elif str(my_url).endswith('.xsd'):
                self.__xsd_doms[str(my_url)] = dom
            else:
                raise WsdlException, 'import error unknown file type (%s)' %(str(my_url))

            imports = xpath.Evaluate('import[@location]', dom.documentElement)
            base_uri = str(my_url)[:str(my_url).rfind('/')+1]
            for node in imports:
                file_path = xpath.Evaluate('string(@location)', node)
                import_url = Url(base_uri+file_path)
                doc_node = self.__readFromUrl(reader, str(import_url))
                dom.documentElement.appendChild(doc_node.documentElement)

        return dom


###########################
# Markers
##########################
class SoapBindingElement(XmlBase, FromDom): pass

class WsdlComponent(XmlBase, FromDom):
    """
    """
    def __init__(self):
        self.name = None

    def _checkNode(self, tagName, node):
        if SplitQName(node.tagName)[1] != tagName:
            raise WsdlException, 'node<%s> is not <%s>' %(node.tagName, tagName)
        
    def setName(self, node):
        self.name = xpath.Evaluate('string(@name)', node)

class DescriptionComponent(WsdlComponent):
    """
        ServiceType,Service,Binding,PortType,Message,Type
    """
    def __init__(self):
        WsdlComponent.__init__(self)

class InformationItem(WsdlComponent):
    """
        Port,Operation,Input,Output,Fault
    """
    def __init__(self):
        WsdlComponent.__init__(self)
        
class OperationInformationItem(InformationItem):
    """
        property:
            message - qname
    """
    def __init__(self):
        self.message = MessageDescriptionComponent()
        return
    
    def __setMessage(self, node):
        self.message.fromDom(node)
        return
    
    def fromDom(self, node):
        self.__setMessage(node)
        return
    
###########################
##########################
class ServiceType(DescriptionComponent):
    """
        Defines grouping of a set of portTypes
        property:
            name*
            targetNamespace
        components:
            portType - qname
    """ 
    def __init__(self): pass

###########################
##########################
class Service(DescriptionComponent):
    """
        User Interface, from here all information. NO LOOKUP UPS.

        Defines a Web service, a collection of ports.
        property:
            name* - w/targetNamespace = qname
            serviceType* - qname of serviceType
            targetNamespace
        components:
            port - contain 1+

	STRICT: require serviceType
            Wsdl1.2 serviceType is required 
            serviceType should contain port.name
    """
    def __init__(self):
        DescriptionComponent.__init__(self)
        self.targetNamespace = None
        self.ports = []
        return

    def fromDom(self, node):
        self.setName(node)
        self.targetNamespace = xpath.Evaluate('string(@targetNamespace)', node.parentNode)

        qname = xpath.Evaluate('string(@serviceType)', node)
        if qname:
	    print_debug('Service.fromDom, not checking serviceType')

        pnodes = xpath.Evaluate('port', node)
        for i in range(len(pnodes)):
            self.ports.append(Port())
            self.ports[i].fromDom(pnodes[i])
        return

###########################
##########################
class SoapBinding(SoapBindingElement):
    """
        style     - rpc/document
        transport - 
    """
    def fromDom(self, node):
        if node.tagName != 'soap:binding':
            raise WsdlException, 'node (%s) not a soap:binding' %(node.tagName)
        self.style = xpath.Evaluate('string(@style)', node)
        self.transport = xpath.Evaluate('string(@transport)', node)
        return
    
class SoapOperation(SoapBindingElement):
    """
        soapAction 
    """
    def fromDom(self, node):
        if node.tagName != 'soap:operation':
            raise WsdlException, 'node (%s) not a soap:operation' %(node.tagName)
        self.soapAction = xpath.Evaluate('string(@soapAction)', node)
        return        

class SoapBody(SoapBindingElement):
    """
        use
        namespace
        encodingStyle
    """
    def fromDom(self, node):
        if node.tagName != 'soap:body':
            raise WsdlException, 'node (%s) not a soap:body' %(node.tagName)
        self.use = xpath.Evaluate('string(@use)', node)
        self.namespace = xpath.Evaluate('string(@namespace)', node)
        self.encodingStyle = xpath.Evaluate('string(@encodingStyle)', node)
        return
    
class SoapAddress(SoapBindingElement):
    """
        location 
    """
    def fromDom(self, node):
        if node.tagName != 'soap:address':
            raise WsdlException, 'node (%s) not a soap:address' %(node.tagName)
        self.location = xpath.Evaluate('string(@location)', node)
        return

class SoapFault(SoapBindingElement):
    """
        name
        use
    """
    def fromDom(self, node):
        if node.tagName != 'soap:fault':
            raise WsdlException, 'node (%s) not a soap:fault' %(node.tagName)
        self.name = xpath.Evaluate('string(@name)', node)
        self.use = xpath.Evaluate('string(@use)', node)
        return
    
class SoapBindingContainer:
    """
        methods for grabbing soap:* tags
    """
    tags = {'soap:body':(xpath.Compile('soap:body'),SoapBody),\
            'soap:binding':(xpath.Compile('soap:binding'),SoapBinding),\
            'soap:operation':(xpath.Compile('soap:operation'),SoapOperation),\
            'soap:address':(xpath.Compile('soap:address'),SoapAddress),\
            'soap:fault':(xpath.Compile('soap:fault'),SoapFault)}
    def __init__(self):
        self.soap = {}
        return

    def __setSOAP(self, node, tag, **xmlns):
        con = xpath.Context.Context(node, processorNss=xmlns)
        snodes = (SoapBindingContainer.tags[tag][0]).evaluate(con)
        
        print_debug('setSOAP XMLNS %s' %(xmlns))
        if len(snodes) > 1:
            raise WsdlException, '(%d) %s children' %(len(snodes),tag)
        elif len(snodes):
            self.soap[tag] = SoapBindingContainer.tags[tag][1]()
            self.soap[tag].fromDom(snodes[0])
        return
    
    def setSOAPBody(self, node, **xmlns):
        self.__setSOAP(node, 'soap:body', **xmlns)
    
    def setSOAPBinding(self, node, **xmlns):
        self.__setSOAP(node, 'soap:binding', **xmlns)
    
    def setSOAPOperation(self, node, **xmlns):
        self.__setSOAP(node, 'soap:operation', **xmlns)
    
    def setSOAPAddress(self, node, **xmlns):
        self.__setSOAP(node, 'soap:address', **xmlns)

    def setSOAPFault(self, node, **xmlns):
        self.__setSOAP(node, 'soap:fault', **xmlns)


###########################
##########################        

class Port(InformationItem, SoapBindingContainer):
    """
        Defines a Web service, a collection of ports.
        property:
            name* - qname w/i enclosing wsdl document
            binding 
        components:
            port - contain 1+
    """
    def __init__(self):
        InformationItem.__init__(self)
        SoapBindingContainer.__init__(self)
        self.binding = Binding()
        return
        
    def fromDom(self, node): 
        self.setName(node)
        binding = xpath.Evaluate('string(@binding)', node)

        lookup = DomLookUp()
        lookup.setDefinitionsNode(node)
        self.xmlns = lookup.xmlns

        self.binding.fromDom(lookup.getBinding(binding))
        self.setSOAPAddress(node, **self.xmlns)
        return

    
class Binding(DescriptionComponent, SoapBindingContainer):
    """
        Framework for binding details for a portType
        property:
            name* - w/targetNamespace = qname
            type* - qname of portType
            targetNamespace
        components:
            soap:binding
            operation
        
    """
    def __init__(self):
        DescriptionComponent.__init__(self)
        SoapBindingContainer.__init__(self)
        self.type = PortType()
        self.operations = {}
        return
    
    def fromDom(self, node):
        self.setName(node)

        lookup = DomLookUp()
        lookup.setDefinitionsNode(node)
        self.xmlns = lookup.xmlns

        portType = xpath.Evaluate('string(@type)', node)
        self.type.fromDom(lookup.getPortType(portType))

        operations = xpath.Evaluate('operation', node)
        for opnode in operations:
            bop = BOperation()
            bop.fromDom(opnode, **self.xmlns)
            self.operations[bop.name] = bop

        self.setSOAPBinding(node, **self.xmlns)
        return


class BindingInformationItem(InformationItem, SoapBindingContainer):
    def __init__(self):
        InformationItem.__init__(self)
        SoapBindingContainer.__init__(self)
        return
      
class BOperation(BindingInformationItem):
    def __init__(self):
        BindingInformationItem.__init__(self)
        self.input  = None
        self.output = None
        self.faults = []
        return
    
    def fromDom(self, node, **xmlns):
        self.setName(node)
        self.setSOAPOperation(node, **xmlns)
        print_debug("Operation: %s" %(node))   
        inodes = xpath.Evaluate('input', node)
        for i in inodes:
            if self.input:
                raise WsdlException, 'Operation (%s) 0 or 1 input (provided %d)'\
                      %(self.name,len(inodes))
            self.input = BInput()
            self.input.fromDom(i, **xmlns)    
        onodes = xpath.Evaluate('output', node)
        for o in onodes:
            if self.output:
                raise WsdlException, 'Operation (%s) 0 or 1 output (provided %d)'\
                      %(self.name,len(onodes))
            self.output = BOutput()
            self.output.fromDom(o, **xmlns)
        fnodes = xpath.Evaluate('fault', node)
        for f in fnodes:
            fault = BFault()
            fault.fromDom(f, **xmlns)
            self.faults.append(fault)
        return
    
class BInput(BindingInformationItem):
    def __init__(self):
        BindingInformationItem.__init__(self)

    def fromDom(self, node, **xmlns):
        self.setSOAPBody(node, **xmlns)
        return
    
class BOutput(BindingInformationItem):
    def __init__(self):
        BindingInformationItem.__init__(self)

    def fromDom(self, node, **xmlns):
        self.setSOAPBody(node, **xmlns)
        return
    
class BFault(BindingInformationItem):
    def __init__(self):
        BindingInformationItem.__init__(self)

    def fromDom(self, node, **xmlns):
        self.setSOAPFault(node, **xmlns)
        return
    
###########################
##########################
class PortType(DescriptionComponent):
    """
        Named set of operations.
        property:
            name - http://www.w3.org/2002/07/wsdl
        components:
            operation
            
    """
    def __init__(self):
        DescriptionComponent.__init__(self)
        self.operations = []
        return
    
    def fromDom(self, node):
        print_debug("PortType: %s" %(node))
        self.setName(node)
        operations = xpath.Evaluate('operation', node)
        for opnode in operations:
            op = Operation()
            op.fromDom(opnode)
            self.operations.append(op)
        return

###########################
# Information Items
##########################
class Operation(InformationItem):
    """
        property:
            name   - http://www.w3.org/2002/07/wsdl
            input  -
            output -
            faults - list
    """
    def __init__(self):
        InformationItem.__init__(self)
        self.input  = None
        self.output = None
        self.faults = []
        return
    
    def fromDom(self, node):
        self.setName(node)
        print_debug("Operation: %s" %(node))   
        inodes = xpath.Evaluate('input', node)
        for i in inodes:
            if self.input:
                raise WsdlException, 'Operation (%s) 0 or 1 input (provided %d)'\
                      %(self.name,len(inodes))
            self.input = Input()
            self.input.fromDom(i)    
        onodes = xpath.Evaluate('output', node)
        for o in onodes:
            if self.output:
                raise WsdlException, 'Operation (%s) 0 or 1 output (provided %d)'\
                      %(self.name,len(onodes))
            self.output = Output()
            self.output.fromDom(o)
        fnodes = xpath.Evaluate('fault', node)
        for f in fnodes:
            fault = Fault()
            fault.fromDom(f)
            self.faults.append(fault)
        return
    
class Input(OperationInformationItem):
    def __init__(self):
        OperationInformationItem.__init__(self)
        return

class Output(OperationInformationItem):
    def __init__(self):
        OperationInformationItem.__init__(self)
        return
    
class Fault(OperationInformationItem):
    """
        property:
            message
            name
    """
    def __init__(self):
        InformationItem.__init__(self)
        OperationInformationItem.__init__(self)
        return
    
    def fromDom(self, node):
        self.setName(node)
        OperationInformationItem.fromDom(self, node)
        return
    
    
######################################################
# MessageDescriptionComponent
#   -defined by Message
####################################################
class Message(DescriptionComponent):
    """
        message name
            parts
    """
    def __init__(self):
        self.parts = []

    def fromDom(self, node):
        self.setName(node)
        parts =  xpath.Evaluate('part', node)
        for part in parts:
            p = Part()
            p.fromDom(part)
            self.parts.append(p)
        return
   
 
class MessageDescriptionComponent(Message):
    """
        fromDom - [input,output,fault] node
          definition-service-port-binding-portType-operation-[input,output,fault]
        type 
    """
    def __init__(self):
        Message.__init__(self)
        
    def fromDom(self, node):
        print_debug("MessageDescriptionComponent: %s" %(node))
        message = xpath.Evaluate('string(@message)', node)
        lookup = DomLookUp()
        lookup.setDefinitionsNode(node)
        mnode = lookup.getMessage(message)
        if mnode:
            Message.fromDom(self, mnode)
        else:
            raise WsdlException, 'Message (%s) node not found' %(message)
        return

class Part(InformationItem):
    """
        definition--message--part
           name
           element --> TypeDescriptionComponent
           type    --> TypeDescriptionComponent
    """
    def __init__(self):
        self.type = None
        self.element = None
        return
    
    def __setType(self, node):
        self.type = xpath.Evaluate('string(@type)', node)
        print_debug('Part.__setType (%s)' %(self.type))
        if self.type:
            tdc = TypeDescriptionComponent(self.type)
            tdc.fromDom(node)
            #self.type = tdc.type
            self.type = tdc
        return

    def __setElement(self, node):
        qname = xpath.Evaluate('string(@element)', node)
        print_debug('Part.__setElement (%s)' %(qname))
        if qname:
            tdc = TypeDescriptionComponent(qname)
            tdc.fromDom(node)
            #self.element = tdc.type
            self.element = tdc
        return

    def fromDom(self, node):
        self.setName(node)
        try:
            self.__setType(node)
            self.__setElement(node)
        except NoItemException, ex:
            print_debug('NoItemException caught %s' %(ex.args))
        return
   

class NoItemException(WsdlException): pass

class Types(DescriptionComponent):
    """ Contains a <schema> node

        Schema defines own xmlns
        gt3 factory:
	   WRONG? <definitions xmlns><types><xsd:schema xmlns:xsd> 
	   CORRECT: <definitions xmlns:xsd><types><xsd:schema xmlns:xsd> 
          
	
    """
    def __init__(self):
        self.schema = Schema()
            
    def fromDom(self, node):
        self._checkNode('types', node)
        
        nodes = []
        for c in node.childNodes:
            if c.nodeType == Node.ELEMENT_NODE:
                nodes.append(c)
        
        if len(nodes) > 1 or not nodes:
            raise WsdlException, '<types> node must have 1 <schema> child, contains' %(len(nodes))
        self.schema.fromDom(nodes[0])
        return
       

######################################################
# Url
#
####################################################
class UrlException(Exception): pass
class Url:
    """
        Wsdl documents may import docs that have already been
        grabbed.
    """
    def __init__(self, url):
        self.__scheme, self.__netloc, self.__path, self.__parameters, \
        self.__query, self.__fragment = urlparse.urlparse(url)
        if not self.__path: 
            UrlException, 'no path in (%s)' %(url)
        self.__condensePath()
        return

    def __condensePath(self):
        tok_idx = self.__path.rfind('/../')
        rear = ''
        while tok_idx > 0:
            skip = 1
            rear = self.__path[tok_idx+4:] + rear
            while self.__path[:tok_idx].endswith('/..'):
                tok_idx -= 3
                skip += 1
            bs_idx = tok_idx
            while skip:
                bs_idx = self.__path[:bs_idx].rfind('/')
                skip -= 1
            self.__path = self.__path[:bs_idx+1]
            tok_idx = self.__path.rfind('/../')
        self.__path += rear
        return

    def __eq__(self, other):
        return str(self) == str(other)
           
    def __str__(self):
        return urlparse.urlunparse((self.__scheme,\
                                   self.__netloc,\
                                   self.__path,\
                                   self.__parameters,\
                                   self.__query,\
                                   self.__fragment))
