############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import types, sys
from xml.dom.ext import SplitQName
from pyGridWare.typecodes import BaseTypeInterpreter
from pyGridWare.schema import *

"""
wsdl2python:
    This module generates an client interface module and a module containing
    all of the types (typecodes) for use with the zsi library.  

Arch:
    WriteServiceModule
      -Takes a wsdl object and creates the client interface and typecodes for 
       the service.
      -Handles writing of import complexities and class definition order.
      -Delegates to ServiceDescription interpretation of service definition.
      -Delegates to SchemaDescription interpretation of schema definition.

    ServiceDescription
      -Interprets the service definition, creates client interface and
       port descriptions code.
      -Delegates to MessageWriter

       MessageWriter
         -Generates a message's class definition.

          PartWriter
            -Generates a message part's description.


    SchemaDescription
      -Interprets the schema definition, creates typecode module code.

       TypeWriter
         -Generates a type's description/typecode.
"""


ID1 = '    '
ID2 = 2*ID1
ID3 = 3*ID1
ID4 = 4*ID1
ID5 = 5*ID1
ID6 = 6*ID1

DEBUG  = 6
NSDICT = {}

def print_debug(msg, level=1, *l, **kw):
    if DEBUG >= level:
        for i in l:
            msg += '\n*\t' + str(i)
        for k,v in kw.items():
            msg += '\n**\t%s: %s' %(k,v)
        print "WSDL2PYTHON(%d): %s" %(level,msg)

class WriteServiceModule:
    """Takes a wsdl object and creates the client interface and typecodes for 
       the service.
    """
    def __init__(self, wsdl):
        """
            wsdl - wsdl object
        """
        self.wsdl = wsdl

    def write(self):
        self.tns_wrote = {}

        f_types = '%s_services_types.py' %(SplitQName(self.wsdl.name)[1])
        header = '%s \n# %s.py \n# generated by %s \n# \n# \n%s\n\n'\
	          %('#'*50, f_types, self.__module__, '#'*50)

        imports = ['\nimport ZSI']
        #imports.append('\nfrom pyGridWare.typecodes import TC_Struct as Struct')
        imports.append('\nfrom ZSI.TCcompound import Struct')
        
        fd = open(f_types, 'w+')
	fd.write(header)
	fd.writelines(imports)
	fd.write('\n'*2)
        for tns, schema in self.wsdl.schemas.items():
            self.write_dependent_schema(tns, schema, fd)
	fd.close()

        f_services = '%s_services.py' %(SplitQName(self.wsdl.name)[1])
        header = '%s \n# %s.py \n# generated by %s \n# \n# \n%s\n\n'\
	          %('#'*50, f_services, self.__module__, '#'*50)

        fd = open(f_services, 'w+')
	fd.write(header)
        for service in self.wsdl.services:
            sd = ServiceDescription()
            sd.imports += ['\nfrom %s import *' %f_types.split('.')[0]]
            sd.fromWsdl(service)
            sd.write(fd)
	fd.close()

    def write_dependent_schema(self, tns, schema, fd):
        """ Not guaranteed to work for mutual depenedencies
        """
        for ns in schema.xmlns.values():
            if self.wsdl.schemas.has_key(ns) and\
               (not self.tns_wrote.has_key(ns)) and\
               (ns != schema.targetNamespace):
                self.write_dependent_schema(ns, self.wsdl.schemas[ns], fd)
        else:
            if not self.tns_wrote.has_key(schema.targetNamespace):
                self.tns_wrote[schema.targetNamespace] = 1
                sd = SchemaDescription()
                sd.fromWsdl(schema)
                sd.write(fd)
 
class ServiceDescription:
    def __init__(self):
	self.imports = []
        return

    def write(self, fd=sys.stdout):
	fd.writelines(self.imports)
	fd.write(self.serviceInterface)
	fd.write(self.serviceLocator)
        fd.write(self.serviceBindings)
        for k, msg in self.messages.items():
            fd.write(str(msg))
            
    def fromWsdl(self, service):
	self.imports += '\nimport urlparse'
	self.imports += '\nfrom pyGridWare.stub import Service, Stub'
	self.imports += '\nfrom ZSI.TCcompound import Struct'
	self.imports += '\nimport ZSI'
        self.messages = {}
	self.serviceInterface = '\n\nclass %sInterface(Service):\n' %(service.name)
	self.serviceLocator = '\n\nclass %sLocator(%sInterface):\n' %(service.name, service.name)
        self.serviceBindings = ''
        for p in service.ports:
            # -----------------------------
	    #  port.soap
	    #    soap:address
            #       location
            # -----------------------------

            # -----------------------------
	    #  class variable address
	    self.serviceLocator += ID1 
	    self.serviceLocator += '%s_address = "%s"\n' %(p.binding.type.name, p.soap['soap:address'].location)

            # -----------------------------
	    #  getPortTypeAddress
	    self.serviceLocator += ID1 
	    self.serviceLocator +=  'def get%sAddress(self):\n' %(p.binding.type.name)
	    self.serviceLocator += ID2 
	    self.serviceLocator +=  'return %sLocator.%s_address\n\n' %(service.name, p.binding.type.name)

            # -----------------------------
            #  getPortType(portAddress)
	    self.serviceInterface += ID1 + 'def get%s(self, portAddress=None, **kw):\n' %(p.binding.type.name)
	    self.serviceInterface += ID2 
	    self.serviceInterface +=  'raise NonImplementationError, "method not implemented"\n'

	    self.serviceLocator += ID1 + 'def get%s(self, portAddress=None, **kw):\n' %(p.binding.type.name)
	    self.serviceLocator += ID2 
	    self.serviceLocator +=  'return %sSOAP(portAddress or %sLocator.%s_address, **kw)\n'\
	                            %(p.binding.name, service.name, p.binding.type.name)

	    myBinding = {'class':'\n\nclass %sSOAP(Stub):' %(p.binding.name), 'defs':{} }

            # -----------------------------
            # Methods
            # -----------------------------
            myBinding['defs']['__init__'] =  '\n%sdef __init__(self, addr, **kw):' %(ID1)
            myBinding['defs']['__init__'] += '\n%snetloc = (urlparse.urlparse(addr)[1]).split(":") + [80,]' %(ID2)
            myBinding['defs']['__init__'] += '\n%skw.update({"host":netloc[0], "port":(int(netloc[1]))})' %(ID2)
            myBinding['defs']['__init__'] += '\n%sStub.__init__(self, **kw)' %(ID2)

            for op in p.binding.type.operations:
                # -----------------------------
                #  REQUIREMENTS SOAP Bindings
	        #  port.binding.operations.soap[binding,operations]
	        #  port.binding.soap[binding]
                #        style, transport
                # -----------------------------
                transport = getattr(p.binding.operations[op.name].soap.get('soap:binding'), 'transport', None) or\
                            p.binding.soap['soap:binding'].transport

                style     = getattr(p.binding.operations[op.name].soap.get('soap:binding'), 'style', None) or\
                            getattr(p.binding.operations[op.name].soap.get('soap:operation'), 'style', None) or\
                            p.binding.soap['soap:binding'].style

                soapAction = getattr(p.binding.operations[op.name].soap.get('soap:operation'), 'soapAction', None)

                encodingStyle = p.binding.operations[op.name].input.soap['soap:body'].encodingStyle
                use = p.binding.operations[op.name].input.soap['soap:body'].use
                namespace = p.binding.operations[op.name].input.soap['soap:body'].namespace

                # -----------------------------
		# Methods
                # -----------------------------
		if op.input:
		    myBinding['defs'][op.name] = '\n%sdef %s(self, request):' %(ID1, op.name)

		    myBinding['defs'][op.name] +=\
                        '\n%sif not isinstance(request, %s):' %(ID2, op.input.message.name)
        
		    myBinding['defs'][op.name] +=\
		        '\n%sraise TypeError, %s' %(ID3, r'"%s incorrect request type" %(request.__class__)')

		    myBinding['defs'][op.name] +=\
                        '\n%sbindings = {"url":"%s",\\\n%s"soap_transport":"%s"}'\
                            %(ID2, soapAction, ID5, transport)

                    self.messages[op.input.message.name] = self.__class__.MessageWriter()

                    if style == 'rpc' and use == 'encoded':
                        self.messages[op.input.message.name].fromMessageEncoded(op.input.message, op.name, namespace)
                    elif style == 'document' and use == 'literal':
                        self.messages[op.input.message.name].fromMessageLiteral(op.input.message, op.input.message.name, namespace)
                    else:
                        raise RuntimeError, 'unsupported combination use=%s style=%s' %(use, style)

		    if op.output:
			myBinding['defs'][op.name] +=\
                            '\n%sresponse = Stub.RPC(self, request, %s(), **bindings)' %(ID2, op.output.message.name)
			myBinding['defs'][op.name] +=\
			    '\n%sif not isinstance(response, %s):' %(ID2, op.output.message.name)
			myBinding['defs'][op.name] +=\
			    '\n%sraise TypeError, %s' %(ID3, r'"%s incorrect response type" %(response.__class__)')
		        myBinding['defs'][op.name] += '\n%sreturn response' %(ID2)
                        self.messages[op.output.message.name] = self.__class__.MessageWriter()
                        
                        if style == 'rpc' and use == 'encoded':
                            self.messages[op.output.message.name].fromMessageEncoded(op.output.message, op.output.message.name, namespace)
                        elif style == 'document' and use == 'literal':
                            self.messages[op.output.message.name].fromMessageLiteral(op.output.message, op.output.message.name, namespace)
                        else:
                            raise RuntimeError, 'unsupported combination use=%s style=%s' %(use, style)

		    else:
			myBinding['defs'][op.name] += '\n%sStub.Send(self, request)' %(ID2)
		        myBinding['defs'][op.name] += '\n%sreturn' %(ID2)

		elif op.output:
		    #NEED TO HAVE A SPECIAL TC CHOICE OBJECT FOR ASYNCH COMMUNICATION
		    pass
		else:
		    raise RuntimeError, 'Operation w/o input and/or output'

            else:
	        self.serviceBindings += myBinding['class']
		self.serviceBindings += "\n%s\n" %(myBinding['defs']['__init__'])
		del myBinding['defs']['__init__']
		for mn,d in myBinding['defs'].items():
		    self.serviceBindings += "\n%s\n" %(d)
		else:
	            self.serviceBindings += "\n"
        
        return

    class MessageWriter:
	def __init__(self):
	    self.typecode = None

	def __str__(self):
	    return self.typecode

	def fromMessageEncoded(self, message, name, namespace):
	    l = []

	    self.typecode = '\n\nclass %s: ' %(message.name)
	    self.typecode += '\n%sdef __init__(self):' %(ID1)
	    for p in message.parts:
                if hasattr(p, 'type'):
                    tp = self.__class__.PartWriter()
                    tp.fromPart(p)
                    l += tp.typecode
                elif hasattr(p, 'element'):
                    raise RuntimeError, 'Bad encoding'
                else:
                    raise RuntimeError, 'FIX'
	    else:
		tcs = ''
		for i in l: tcs += (i + ',')
		self.typecode += '\n%s%s.typecode = Struct(%s,[%s],pname="%s",aname="_%s",oname="%s  xmlns=\\"%s\\"")'\
			     %(ID2,message.name,message.name,tcs,name,name,name,namespace)


	def fromMessageLiteral(self, message, name, namespace):
	    l = []

	    self.typecode = '\n\nclass %s: ' %(message.name)
	    self.typecode += '\n%sdef __init__(self):' %(ID1)
	    for p in message.parts:
		if p.element:
		    tp = self.__class__.PartWriter()
		    tp.fromPart(p)
		    if tp.name:
			self.typecode = '\n\nclass %s(%s): ' %(message.name, tp.name)
			self.typecode += '\n%sdef __init__(self, name=None):' %(ID1)
			self.typecode += '\n%s%s.__init__(self)' %(ID2, tp.name)
		    else:
			self.typecode = '\n\nclass %s: ' %(message.name)
			self.typecode += '\n%sdef __init__(self, name=None): pass' %(ID1)
		    break
		else:
		    if hasattr(p, 'type'):
			tp = self.__class__.PartWriter()
			tp.fromPart(p)
			l += tp.typecode
		    else:
			raise RuntimeError, 'Missing attribute for <message name=%s><part name=%s>' %(message.name,p.name)

	    else:
		tcs = ''
		for i in l: tcs += (i + ',')
		self.typecode += '\n%s%s.typecode = Struct(%s,[%s],pname="%s",aname="_%s",oname="%s  xmlns=\\"%s\\"")'\
			     %(ID2,message.name,message.name,tcs,message.name,message.name,message.name,namespace)

	class PartWriter:
	    def __init__(self):
		self.typecode = None
		#self.element = None
		self.name = None

	    def __recurse_tdc(self, tp):
		tp.type
		if isinstance(tp.type, TypeDescriptionComponent):
		    tp.type
		    tp = self.__recurse_tdc(tp.type)
		else:
		    return tp.type

	    def fromPart(self, part):
		bti = BaseTypeInterpreter()

		if part.type:
		    tp = self.__recurse_tdc(part.type)
		else:
		    tp = self.__recurse_tdc(part.element)
		    self.name = tp.name
		    return

		self.typecode = []

		if isinstance(tp, types.StringType) or isinstance(tp, types.UnicodeType):
		    raise RuntimeError, 'shouldnt happen'

		elif isinstance(tp, Element):
		    tpe = self.__recurse_tdc(tp.type)
		    if isinstance(tpe, types.StringType) or isinstance(tpe, types.UnicodeType):
			tpc = bti.get_typeclass(tpe)
			self.typecode.append('%s(pname="%s",aname="_%s",optional=1)' %(tpc, part.name, part.name))
		    elif isinstance(tpe, ComplexType):
			self.typecode.append('%s("%s").typecode' %(tpe.name, part.name))

		elif isinstance(tp, SimpleType):
		    if tp.qname:
			tpc = bti.get_typeclass(tp.qname, tp.targetNamespace)
			self.typecode.append('%s(pname="%s",aname="_%s",optional=1)' %(tpc, part.name, part.name))
		    elif tp.name:
			self.typecode.append('%s(pname="%s",aname="_%s",optional=1)' %(tp.name, part.name, part.name))
		    else:
			raise RuntimeError, 'shouldnt happen'

		elif isinstance(tp, ComplexType):
		    self.typecode.append('Struct(%s, %s().definition, pname="%s",aname="_%s",optional=1)' %(tp.name, tp.name, part.name, part.name))

		else:
		    self.typecode.append('ZSI.TC.XML()')


		return


class SchemaDescription:
    def __init__(self):
        #self.imports = []
        return

    def fromWsdl(self, schema):
        if not isinstance(schema, Schema):
	    raise TypeError, 'type %s not a Schema' %(schema.__class__)

	self.header = '%s \n# %s \n#\n# %s \n%s\n'\
                      %('#'*30, 'targetNamespace', schema.targetNamespace, '#'*30)
	self.body = ''
        class_dict = {}
        class_list = []

	for name, tp in schema.types.items():
            tw = self.__class__.TypeWriter()
            tw.fromType(tp)
            if tw.precede:
                if class_dict.has_key(tw.precede):
                    class_dict[tw.precede].append(tw)
                else:
                    class_dict[tw.precede] = [tw]
            else:
                class_list.append(tw.name)
                self.body += tw.classdef
                self.body += tw.initdef
        else:
            self.getClassDefs(class_list, class_dict)

    def getClassDefs(self, class_list, class_dict):
        check_list = []
        for indx in range(len(class_list)):
            if class_dict.has_key(class_list[indx]):
                for tw in class_dict[class_list[indx]]:
                    self.body += tw.classdef
                    self.body += tw.initdef
                    check_list.append(tw.name)
                else:
                    del class_dict[class_list[indx]]
        if check_list:
            self.getClassDefs(check_list, class_dict)
        else:
            for l in class_dict.values(): 
                for tw in l:
                    self.body += tw.classdef
                    self.body += tw.initdef


    def write(self, fd=sys.stdout):
	fd.write(self.header)
	fd.write(self.body)
	fd.write('\n'*4)

    class TypeWriter:
	def __init__(self):
	    self.bti = BaseTypeInterpreter()
	    self.name = None
	    self.precede = None
	    self.initdef  = None
	    self.classdef = None
	    return

	def fromType(self, myType):
	    tp   = myType
	    if isinstance(tp, TypeDescriptionComponent):
		tp = self.__recurse_tdc(tp)

	    self.name = tp.name
	    if isinstance(tp, SimpleType):
		if tp.name:
		    tpbase = self.__recurse_tdc(tp.restriction.base)
		    tpc = self.bti.get_typeclass(tpbase.qname, tpbase.targetNamespace)
		    self.initdef  = '\n%sdef __init__(self,pname,optional=1,repeatable=0):' %(ID1)
		    if tpc:
			self.precede  = '%s' %(tpc)
			self.classdef = '\nclass %s(%s):' %(tp.name, tpc)
			self.initdef  += '\n%s%s.__init__(self,pname="%s",optional=optional,repeatable=repeatable)'\
					 %(ID2,tpc,tp.name)
		    else:
			self.classdef = '\nclass %s(ZSI.TC.Any):' %(tp.name)
			self.initdef  += '\n%sZSI.TC.Any.__init__(self,pname,optional=optional,repeatable=repeatable)'\
					 %(ID2)
		else:
		    raise RuntimeError, 'shouldnt happen'

	    elif isinstance(tp, WildCardDeclaration):
		self.classdef = '\nclass %s(ZSI.TC.Any):' %(tp.name)
		self.initdef  = '\n%s__init__(self,pname):' %(ID1)
		self.initdef  += '\n%sZSI.TC.Any.__init__(self,pname)' %(ID2)

	    elif isinstance(tp, Element):
		etp = tp.type
		if isinstance(etp, TypeDescriptionComponent):
		    etp = self.__recurse_tdc(tp.type)
		    
		if isinstance(etp, SimpleType):
		    tpc = self.bti.get_typeclass(tp.name, tp.targetNamespace)
		    self.precede  = '%s' %(tpc)
		    self.classdef = '\nclass %s(%s):' %(tp.name, tpc)
		    self.initdef  = '\n%sdef __init__(self, name=None):' %(ID1)
		    self.initdef  += '\n%s%s.__init__(self,pname="%s")' %(ID2,tpc,tp.name)

		elif isinstance(etp, ComplexType):
		    if etp.name:
			#Global
			self.precede  = '%s' %(etp.name)
			self.classdef = '\n\nclass %s(%s):' %(tp.name, etp.name)
			self.initdef  = '\n%sdef __init__(self, name="%s"):' %(ID1, tp.name)
			self.initdef += '\n%s%s.__init__(self)' %(ID2, etp.name)
			self.initdef += '\n%s%s.typecode = Struct(%s,%s().definition,pname="%s",aname="_%s",oname="%s  xmlns=\\\"%s\\\"")'\
			    %(ID2,tp.name,tp.name,etp.name,tp.name,tp.name,tp.name,tp.targetNamespace)
		    else:
			#Local: TODO look w/i complexType tag
			self.classdef = '\n\nclass %s:' %(tp.name)
			self.initdef  = '\n%sdef __init__(self): pass' %(ID1)
			self.initdef += '\n%s.typecode = Struct(%s,[],pname="%s",aname="_%s")'\
			    %(tp.name,tp.name,tp.name,tp.name)

		elif not etp:
		    self.classdef = '\n\nclass %s(Struct):' %(tp.name)
		    self.initdef  = '\n%sdef __init__(self, name=None):' %(ID1)
		    self.initdef += '\n%sStruct.__init__(self,(,),pname="%s",aname="_%s",inline=1)'\
			%(ID2,tp.name,tp.name)
		else:
		    raise RuntimeError, 'Unknown type(%s) not handled ' %(etp.__class__)

	    elif isinstance(tp, ComplexType):
		self.classdef   = '\n\nclass %s:' %(tp.name)
		self.initdef    = ''

		tmp = ''
		for a in tp.attributes:
		    tmp += '("%s","%s"),' %(a.name, a.type.qname)
		if tmp:
		    self.initdef += '\n%sattributes = [%s]' %(ID1, tmp)

		self.initdef    += '\n%sdef __init__(self, name="%s"):' %(ID1, tp.name)
		typecodelist = '['


		#
		# HACK: Place all sequence and all ContentType info
		#       into one Struct.
		#  Limits imposed: Cant repeat a sequence or all.
		#
		contents = []
		for all in tp.alls:
		    contents += all
		for sequence in tp.sequences:
		    contents += sequence

		for e in contents:
		    if isinstance(e, Element):
			if hasattr(e, 'type'):
			    if isinstance(e.type, TypeDescriptionComponent):
				etp = e.type
				etp = self.__recurse_tdc(e.type)
			    else:
				etp = e.type
         
			if hasattr(e, 'name') and e.name:    
			    self.initdef += '\n%sself._%s = None' %(ID2, e.name)

			if   isinstance(e, WildCardDeclaration):
			    occurs = ''
			    if e.maxOccurs == 'unbounded' or int(e.maxOccurs) > 1:
				occurs += ', repeatable=1'
			    if int(e.minOccurs) == 0:
				occurs += ', optional=1'
			    typecodelist  += 'ZSI.TC.Any(pname="%s",aname="_%s"%s), ' %(e.name,e.name,occurs)

			elif isinstance(e, AnyType):
			    typecodelist  += 'ZSI.TC.Any(pname="%s",aname="_%s"), ' %(e.name,e.name)

			elif isinstance(etp, SimpleType):
			    occurs = ''
			    if e.maxOccurs == 'unbounded' or int(e.maxOccurs) > 1:
				occurs += ', repeatable=1'
			    if int(e.minOccurs) == 0:
				occurs += ', optional=1'

			    #Base
			    if hasattr(e, 'base'):
				if isinstance(e.base, TypeDescriptionComponent):
				    etp = self.__recurse_tdc(e.base)
				else:
				    etp = e.base

			    tpc = self.bti.get_typeclass(etp.qname, etp.targetNamespace)
			    if tpc:
				typecodelist  += '%s(pname="%s",aname="_%s"%s), ' %(tpc,e.name,e.name,occurs)
			    else:
				typecodelist  += '%s(pname="%s",aname="_%s"%s), ' %(etp.name,e.name,e.name,occurs)

			elif isinstance(etp, ComplexType):
			    if isinstance(etp, LocalElementDeclaration):
			        typecodelist  += 'Struct(%s, %s().definition, pname="%s",aname="_%s"), '\
			                     %(etp.name,etp.name,e.name,e.name,e.name)
                            else:
			        typecodelist  += 'Struct(%s, %s().definition, pname="%s",aname="_%s",oname="%s  xmlns=\\\"%s\\\""), '\
			                     %(etp.name,etp.name,e.name,e.name,e.name,e.targetNamespace)

			elif etp:
			    #self.initdef  += '\n%sself.%s = None' %(ID2, e.name)
			    typecodelist  += '%s(pname="%s",aname="%s"), ' %(etp.name, e.name)
			elif e:
			    raise RuntimeError, 'instance %s not handled ' %(e.type.type)
			else:
			    raise RuntimeError, 'instance %s not handled ' %(e)
		    else:
			raise RuntimeError, 'instance %s not handled ' %(e.__class__)
		else:
		    pass

		for choice in  tp.choices:
		    typecodelist += 'ZSI.TCcompound.Choice(['
		    
		    for e in tp.choice:
			if isinstance(e, Element):
			    if hasattr(e, 'type'):
				if isinstance(e.type, TypeDescriptionComponent):
				    etp = e.type
				    etp = self.__recurse_tdc(e.type)
				else:
				    etp = e.type
				
			    self.initdef += '\n%sself._%s = None' %(ID2, e.name)
			    if isinstance(e, AnyType):
				typecodelist  += 'ZSI.TC.Any(pname="%s",aname="%s"), ' %(e.name,e.name)

			    elif isinstance(etp, SimpleType):
				occurs = ''
				if e.maxOccurs == 'unbounded' or int(e.maxOccurs) > 1:
				    occurs += ', repeatable=1'
				if int(e.minOccurs) == 0:
				    occurs += ', optional=1'
				tpc = self.bti.get_typeclass(etp.qname, etp.targetNamespace)
				typecodelist  += '%s(pname="%s",aname="_%s"%s), ' %(tpc,e.name,e.name,occurs)

			    elif isinstance(etp, ComplexType):
				typecodelist  += 'Struct(%s,%s().typecode,pname="%s",aname="_%s"), '\
					      %(etp.name,etp.name,e.name,e.name)
			    
			    elif etp:
				self.initdef  += '\n%sself._%s = None' %(ID2, e.name)
				typecodelist  += '%s("%s"), ' %(etp.name, e.name)
			    elif e:
				raise RuntimeError, 'instance %s not handled ' %(e.type.type)
			    else:
				raise RuntimeError, 'instance %s not handled ' %(e)
			else:
			    raise RuntimeError, 'instance %s not handled ' %(e.__class__)
		    else:
			if tp.choice:
			    keywords = ''
			    if tp.choice.maxOccurs == 'unbounded' or int(tp.choice.maxOccurs) > 1:
				keywords += '"repeatable":1, '
			    if int(tp.choice.minOccurs) == 0:
				keywords += '"optional":1'
			    typecodelist += '], **{%s})' %(keywords)
		else:
		    pass
		typecodelist += ']'
		self.initdef += '\n%s%s.definition = %s' %(ID2, tp.name, typecodelist)
	    else:
		raise RuntimeError, 'WARNING: NOT HANDLED %s' %(tp.__class__)

	    return


	def __recurse_tdc(self, tp):
	    tp.type
	    if isinstance(tp.type, TypeDescriptionComponent):
		tp.type
		tp = self.__recurse_tdc(tp.type)
	    else:
		return tp.type
