############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
#     xml.xpath.ParsedNodeTest.QualifiedNameTest.match()
#  
###########################################################################

import types
from xml import xpath
from xml.xpath import Conversions
from xml.dom.ext import PrettyPrint, SplitQName
from xml.dom import Node
from threading import RLock
try:
    from ZSI.TC import XSD1
except ImportError, ex:
    XSD1 = 'http://www.w3.org/1999/XMLSchema'


WSDL   = 'http://schemas.xmlsoap.org/wsdl/'
STRICT = 0
DEBUG  = 0
XMLNS = {'xsd':XSD1}


def print_debug(msg, level=1, *l, **kw):
    if DEBUG >= level:
        for i in l:
            msg += '\n*\t' + str(i)
        for k,v in kw.items():
            msg += '\n**\t%s: %s' %(k,v)
        print "XMLINTERFACE(%d): %s" %(level,msg)

class DOMException(Exception): pass

###########################
# Markers
#    common to wsdl and schema
##########################
class XmlBase: 
    """ These class variables are for string indentation.
    """
    __indent = 0
    __rlock = RLock()

    def __str__(self):
        XmlBase.__rlock.acquire()
        XmlBase.__indent += 1
        tmp = "<" + str(self.__class__) + '>\n'
	for k,v in self.__dict__.items():
            tmp += "%s* %s = %s\n" %(XmlBase.__indent*'  ', k, v)
        XmlBase.__indent -= 1
        XmlBase.__rlock.release()
	return tmp

class FromDom:
    def fromDom(self, node):
        raise NotImplementedError, 'fromDom is not implemented in (%s)' %(self.__class__)
    

    
######################################################
#                   DOM
#    import.namespace == definitions.targetNamespace
#
#      dom
#   definitions
#   /    |    \
# services,binding,import -> definitions 
#   /    |     \             /    |    \
# port  operation       services,binding,import -> definitions
####################################################

class DomLookUp:
    """
        Does an evaluate of enclosing dom, then dive into child 
        definitions, which were imported and appended.
 
        Assume only last item is relevant portType[@name="typens:GooglePortType"]
    """
    def __init__(self):
        self._node = None
        self.xmlns = {'wsdl':WSDL}
        self.targetNamespace = None
    
    def __uniqueResult(self, results):
        if results != 1:
            raise DOMException, 'reference must be unique, found (%d) results' %(results)
        
    def __evaluate(self, component, name=None):
        """ Definitions, Schema
        """
        con = xpath.Context.Context(self._node, processorNss=self.xmlns)
        if name:
            exp = xpath.Compile('wsdl:'+component+'[@name="'+name+'"]')
            return exp.evaluate(con) or self.__findNextDefinition(component, name)

        exp = xpath.Compile(component)
        return exp.evaluate(con)

    def _setXmlns(self):
        for a in self._node.attributes:
            if a.name.startswith('xmlns:'):
                prefix, xmlns = SplitQName(a.name)
                self.xmlns[str(prefix)] = a.value
            elif a.name == 'xmlns':
                self.xmlns[str(a.name)] = a.value
            elif a.name == 'targetNamespace':
                self.targetNamespace = a.value
            elif a.name == 'name':
                self.name = a.value
        return
     
    def __findNextDefinition(self, component, qname):
        prefix,name = SplitQName(qname)
        tns = self.xmlns[prefix]
        nodes = None

        if tns == self.targetNamespace:
            nodes = self.__evaluate(component, name)
        else:
            def_eval_str = 'definitions[@targetNamespace="'+tns+'"]'
            nodes = xpath.Evaluate(def_eval_str, self._node)

            for n in nodes:
                lookup = DomLookUp()
                lookup.setDefinitionsNode(n)
                nodes  = lookup.__evaluate(component, name)
                if nodes: 
                    break
            else:
                for i in self._node.childNodes:
                    if i.nodeType == Node.ELEMENT_NODE and i.tagName == 'import':
                        for a in i.attributes:                        
                            if a.name == 'namespace':
                                print_debug("-->%s  tns-%s" %(i.tagName, a.value), 5)
                                break
                    elif i.nodeType == Node.ELEMENT_NODE and i.tagName == 'definitions':
                        for a in i.attributes:                        
                            if   a.name == 'name': v1 = a.value
                            elif a.name == 'targetNamespace':  v2 = a.value
                        else:
                            print_debug("-->%s  name-%s tns-%s" %(i.tagName, v1, v2), 5)
                    elif i.nodeType == Node.ELEMENT_NODE:
                        print_debug("-->%s" %(i.tagName))
                raise RuntimeError, 'couldnt find tns-%s' %(tns)
        return nodes
    
    def setDefinitionsNode(self, node):
        while not node.tagName == 'definitions':
            if node.parentNode:
                node = node.parentNode
            else:
                raise DOMException, 'No parent definitions node'

        self._node = node
        self._setXmlns()

    def setSchemaNode(self, node):
        # Hack to get xsd:schema
        while not (SplitQName(node.tagName)[1] == 'schema'):
            if node.tagName == 'definitions':
                raise DOMException, 'No parent definitions node'
            elif node.parentNode:
                node = node.parentNode
            else:
                raise DOMException, 'No parent schema node'
        self._node = node
        self._setXmlns()
 
    def setSchemaDefNode(self, node):
        # Hack to get xsd:schema
        while not (SplitQName(node.tagName)[1] == 'schema') and\
              not node.tagName == "definitions":
            if node.parentNode:
                node = node.parentNode
            else:
                raise DOMException, 'No parent schema node'
 
        self._node = node
        self._setXmlns()

    def getBinding(self, qname):
        """ returns a dom node <binding>
        """
        print_debug("getBinding: %s" %(qname))
        nodes = self.__evaluate('binding', qname)
        self.__uniqueResult(len(nodes))
        return nodes[0]
    
    def getPortType(self, qname):
        """ returns a dom node <portType>
        """
        print_debug("getPortType: %s" %(qname))
        nodes = self.__evaluate('portType', qname)
        self.__uniqueResult(len(nodes))
        return nodes[0]

    def getMessage(self, qname):
        """ returns a dom node <message>
        """
        print_debug("getMessage: %s" %(qname))
        nodes = self.__evaluate('message', qname)
        self.__uniqueResult(len(nodes))
        return nodes[0]

    def getTypes(self):
        """ returns a list containing <types>
        """
        return xpath.Evaluate('types', self._node)

    def getService(self):
        """ returns a list containing <service>
        """
        return xpath.Evaluate('service', self._node)

    def getImportNS(self):
        ns = {}
        nodes = xpath.Evaluate('import[@namespace]', self._node)    
        for n in nodes:
            ns[n.attributes[('',u'namespace')].value] = n.attributes[('',u'location')].value

        return ns

    def getSchema(self, prefix):
        """ returns a dom node <schema>
            prefix is mapped in local definition to a targetNamespace.
        """
        tns = self.xmlns.get(prefix) or self.targetNamespace
        print_debug('schema  prefix(%s) tns(%s)' %(prefix, tns))
        con = xpath.Context.Context(self._node, processorNss=self.xmlns)
        exp = xpath.Compile('wsdl:schema[@targetNamespace="'+tns+'"]')
        nodes = exp.evaluate(con)
        self.__uniqueResult(len(nodes))
        return nodes[0]

    def getSimpleTypes(self):
        # Hack to get xsd:schema
        return self.__evaluate('simpleType') or self.__evaluate('xsd:simpleType')
        
    def getElements(self):
        # Hack to get xsd:schema
        return self.__evaluate('element') or self.__evaluate('xsd:element')
    
    def getComplexTypes(self):
        # Hack to get xsd:schema
        return self.__evaluate('complexType') or self.__evaluate('xsd:complexType')

    def getTypeDict(self, node):
	types_d = {'element':[], 'complexType':[], 'simpleType':[], 'any':[]}
	# Hack to get xsd:element, etc
        for child in node.childNodes:
            if child.nodeType == Node.ELEMENT_NODE:
	        types_d[(SplitQName(child.tagName)[1])].append(child)

        return types_d
